/******************************************************************************

 @File         OGLESInitialization_SymbianUIQ3.cpp

 @Title        Initialization for Symbian UIQ3 using Khronos EGL

 @Copyright    Copyright (C) 2006 - 2006 by Imagination Technologies Limited.

 @Platform     Symbian UIQ3

 @Description  Basic Tutorial that shows step-by-step how to initialize OpenGL ES
               1.x, use it for clearing the screen with different colours and
               terminate it.

******************************************************************************/
#include "OGLESInitialization.h"
#include <OGLESInitialization.rsg>

/****************************************************************************
 ** DEFINES                                                               **
 ****************************************************************************/

const TUid KUID={UID_OGL};
const TUid KUIDView={0x00442233};


/****************************************************************************
 ** GLOBAL CODE                                                            **
 ****************************************************************************/

/*!***********************************************************************
 @Function		NewApplication
 @description	
*************************************************************************/
LOCAL_C CApaApplication* NewApplication(){
	return new CShellApplication;
}

/*!***********************************************************************
 @Function		E32Main
 @description	Program entry point
*************************************************************************/ 
GLDEF_C TInt E32Main(void){
  return EikStart::RunApplication(NewApplication);
}




/*!***************************************************************************
	Class: CAppView
*****************************************************************************/

/*!***********************************************************************
 @Function		CAppView
 @Input			aAppUi
 @Description	Constructor
*************************************************************************/
CAppView::CAppView(CQikAppUi* aAppUi):CQikViewBase(*aAppUi,KNullViewId)
{
	bQuit = false;
}

/*!***********************************************************************
 @Function		NewL
*************************************************************************/
CAppView* CAppView::NewL(CQikAppUi* aAppUi,const TRect& aRect)
{
	CAppView* cmv = new(ELeave)CAppView(aAppUi);
	CleanupStack::PushL(cmv);
	cmv->ConstructL(aRect);
	CleanupStack::Pop();
	return cmv;
}

/*!***********************************************************************
 @Function		ViewId
 @Description	Returns the app's UID numbers
*************************************************************************/
TVwsViewId CAppView::ViewId(void) const
{
	return TVwsViewId(KUID,KUIDView);
}

/*!***********************************************************************
 @Function		~CAppView
*************************************************************************/
CAppView::~CAppView()
{
  delete iPeriodic;


}

/*!***********************************************************************
 @Function		OfferKeyEventL
 @Description	Implements keyboard response
*************************************************************************/
TKeyResponse CAppView::OfferKeyEventL(const TKeyEvent& aKeyEvent,TEventCode aType)
{
	if(aType == EEventKeyDown)
	{
		RDebug::Printf("Key code: %d",aKeyEvent.iScanCode);
  		switch (aKeyEvent.iScanCode)
		{
  			case 'R':
				break;

			case 'Q': case 172: case 167: case 212:	case 213:
				bQuit = true;
				return EKeyWasConsumed;
		}
	}
	return EKeyWasNotConsumed;
}

/*!***********************************************************************
 @Function		HandleCommandL
 @Description	UIQ command interface
*************************************************************************/
void CAppView::HandleCommandL(CQikCommand& aCommand)
{
	TKeyEvent aKeyEvent;

    switch (aCommand.Id())
    {
        case EQikCmdGoBack:
            RequestFocusL(NULL);

         case EEikCmdExit:
			aKeyEvent.iScanCode = 'Q';
			OfferKeyEventL(aKeyEvent,EEventKeyDown);
			      break;
        default:
            CQikViewBase::HandleCommandL(aCommand);
            break;
    }
}

/*!***********************************************************************
 @Function		ConstructL
 @Description	Second phase constructor
*************************************************************************/
void CAppView::ConstructL(const TRect& /* aRect */)
{
	BaseConstructL();
}

/*!***********************************************************************
 @Function		ViewConstructL
 @Description	
*************************************************************************/
void CAppView::ViewConstructL()
{
	InitComponentArrayL();
	ViewConstructFromResourceL(DSATEST_VIEW_CONFIG,DSATEST_VIEW_CONTROLS);
	
	ActivateL();
	TQikViewMode vm;
	vm.SetFullscreen();

	SetViewModeL(vm);
}

/*!***********************************************************************
 @Function		ViewActivatedL
 @Description	
*************************************************************************/
void CAppView::ViewActivatedL(const TVwsViewId &aPrevViewId, TUid /* aCustomMessageId */, const TDesC8 & /* aCustomMessage */)
{
	SetParentView(aPrevViewId);

	/*
		Initialise
	*/
	// EGL variables
	eglDisplay	= 0;
	eglConfig	= 0;
	eglSurface	= 0;
	eglContext	= 0;

	/*
		Step 1 - Get the default display.
		EGL uses the concept of a "display" which in most environments
		corresponds to a single physical screen. Since we usually want
		to draw to the main screen or only have a single screen to begin
		with, we let EGL pick the default display.
		Querying other displays is platform specific.
	*/
	eglDisplay = eglGetDisplay(EGL_DEFAULT_DISPLAY);

	/*
		Step 2 - Initialize EGL.
		EGL has to be initialized with the display obtained in the
		previous step. We cannot use other EGL functions except
		eglGetDisplay and eglGetError before eglInitialize has been
		called.
		If we're not interested in the EGL version number we can just
		pass NULL for the second and third parameters.
	*/
	EGLint iMajorVersion, iMinorVersion;
	if (!eglInitialize(eglDisplay, &iMajorVersion, &iMinorVersion))
	{
		RDebug::Printf("Error: eglInitialize() failed.");
		delete this;
	}

	/*
		Step 3 - Specify the required configuration attributes.
		An EGL "configuration" describes the pixel format and type of
		surfaces that can be used for drawing.
		For now we just want to use a 16 bit RGB surface that is a 
		Window surface, i.e. it will be visible on screen. The list
		has to contain key/value pairs, terminated with EGL_NONE.
	 */
	EGLint pi32ConfigAttribs[3];
	pi32ConfigAttribs[0] = EGL_SURFACE_TYPE;
	pi32ConfigAttribs[1] = EGL_WINDOW_BIT;
	pi32ConfigAttribs[2] = EGL_NONE;

	/*
		Step 4 - Find a config that matches all requirements.
		eglChooseConfig provides a list of all available configurations
		that meet or exceed the requirements given as the second
		argument. In most cases we just want the first config that meets
		all criteria, so we can limit the number of configs returned to 1.
	*/
	int iConfigs;
	if (!eglChooseConfig(eglDisplay, pi32ConfigAttribs, &eglConfig, 1, &iConfigs) || (iConfigs != 1))
	{
		RDebug::Printf("Error: eglChooseConfig() failed.\n");
		delete this;
	}

	/*
		Step 5 - Create a surface to draw to.
		Use the config picked in the previous step and the native window
		handle when available to create a window surface. A window surface
		is one that will be visible on screen inside the native display (or 
		fullscreen if there is no windowing system).
		Pixmaps and pbuffers are surfaces which only exist in off-screen
		memory.
	*/

	eglSurface = eglCreateWindowSurface(eglDisplay, eglConfig, (NativeWindowType)DrawableWindow(), NULL);
	//if (!TestEGLError("eglCreateWindowSurface"))
	//{
	//	delete this;
	//}

	/*
		Step 6 - Create a context.
		EGL has to create a context for OpenGL ES. Our OpenGL ES resources
		like textures will only be valid inside this context
		(or shared contexts)
	*/
	eglContext = eglCreateContext(eglDisplay, eglConfig, NULL, NULL);
	//if (!TestEGLError("eglCreateContext"))
	//{
	//	delete this;
	//}

	/*
		Step 7 - Bind the context to the current thread and use our 
		window surface for drawing and reading.
		Contexts are bound to a thread. This means you don't have to
		worry about other threads and processes interfering with your
		OpenGL ES application.
		We need to specify a surface that will be the target of all
		subsequent drawing operations, and one that will be the source
		of read operations. They can be the same surface.
	*/
	eglMakeCurrent(eglDisplay, eglSurface, eglSurface, eglContext);
	//if (!TestEGLError("eglMakeCurrent"))
	//{
	//	delete this;
	//}
	

	iPeriodic = CPeriodic::NewL(CActive::EPriorityIdle);
	iPeriodic->Start(100000,10000,TCallBack(CAppView::DrawCallback,this)); 
}

/*!***********************************************************************
 @Function		DrawCallback
 @Description	Called every frame - runs shell code, and user demo
				code.
*************************************************************************/
TInt CAppView::DrawCallback(TAny* aInstance){

	((CAppView*) aInstance)->draw();


	return 0;
}

/*!***********************************************************************
 @Function		draw
 @Description	Called every frame - runs shell code, and user demo
				code.
*************************************************************************/
void CAppView::draw()
{
	static int framenum = 0;
	static float fCol = 0.0f;

	//clear screen to varying colour:
	glClearColor(fCol, fCol, fCol, 1.0f);
	fCol += 0.01f;
	if(fCol >1.0f) fCol -= 1.0f;

	/*
		Clears the color buffer.
		glClear() can also be used to clear the depth or stencil buffer
		(GL_DEPTH_BUFFER_BIT or GL_STENCIL_BUFFER_BIT)
	*/
	glClear(GL_COLOR_BUFFER_BIT);
	//if (!TestEGLError("glClear"))
	//{
	//	Exit();
	//}
	
	/*
		Swap Buffers.
		Brings to the native display the current render surface.
	*/
	eglSwapBuffers(eglDisplay, eglSurface);

	//quit if 'Q' pressed, or frame number passes 640;
	if(bQuit || framenum++ > 640) CBaActiveScheduler::Exit();
}

/*!***********************************************************************
 @Function		ViewDeactivated
 @Description	
*************************************************************************/
void CAppView::ViewDeactivated()
{
	delete iPeriodic;

	eglMakeCurrent(eglDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
	eglTerminate(eglDisplay);
}






/*!***************************************************************************
	Class: CShellAppUI
*****************************************************************************/

/*!***********************************************************************
 @Function		ConstructL
  @Description	Second phase constructor
*************************************************************************/
void CShellAppUI::ConstructL(){
	iCoeEnv->DisableExitChecks(ETrue);

	CQikAppUi::ConstructL();

	AddViewL(*CAppView::NewL(this,ClientRect()));
}

/*!***********************************************************************
 @Function		HandleCommandL
  @Description	Command interface
*************************************************************************/
void CShellAppUI::HandleCommandL(TInt aCommand)
{
  switch (aCommand){
	case EEikCmdExit: 
			CEikAppUi::Exit();
			break;
		default:
			break;
	}
}



/*!***************************************************************************
	Class: CShellDocument
*****************************************************************************/

/*!***********************************************************************
 @Function		CShellDocument
 @Description	Constructor
*************************************************************************/
CShellDocument::CShellDocument(CEikApplication& aApp) : CQikDocument(aApp)
{
	
}

/*!***********************************************************************
 @Function		CShellDocument
 @Description	Destructor
*************************************************************************/
CShellDocument::~CShellDocument()
{
	CloseSTDLIB();
}

/*!***********************************************************************
 @Function		NewL
 @Description	
*************************************************************************/
CShellDocument* CShellDocument::NewL(CEikApplication& aApp)
{
	CShellDocument* self=new(ELeave) CShellDocument(aApp);
	return self;
}

/*!***********************************************************************
 @Function		CreateAppUiL
 @Description	
*************************************************************************/
CQikAppUi* CShellDocument::CreateAppUiL()
{
  return(new(ELeave) CShellAppUI);
}

/*!***********************************************************************
 @Function		AppDllUid
 @Description	
*************************************************************************/
TUid CShellApplication::AppDllUid() const
{
  return(KUID);
}


CApaDocument* CShellApplication::CreateDocumentL()
{
	return CShellDocument::NewL(*this);
}


/*****************************************************************************
 End of file (OGLESInitialization_SymbianUIQ3.cpp)
*****************************************************************************/

